--[[
	This is a XenoBot example script, intended to
	teach new users about the scripting API and 
	act as script that is usable in actual play.
	
	recentLootChannel.lua - highlights lootmessages
	that contain items from a list

	** DO NOT EDIT THIS FILE. INSTEAD, COPY IT TO
	"Documents\XenoBot\Scripts" AND EDIT THE COPY. **
]]--

local items = {"fire sword", "golden armor"} --a list that stores all itemnames that should be highlighted

--this is a function that is triggered on user input
function onSpeak(channel, text) 
	--this says out loud everything you type into the custom-made channel just like any other native channel would do
	Self.Say(text)
end

function onClose(channel)
	--this prints a message into the server log channel to inform you that the custom-made channel was closed
	print(channel .. " was closed.")
end

--this opens a custom-made channel window with the name Recent Loot and the callbacks onSpeak and onClose
local channel = Channel.New("Recent Loot", onSpeak, onClose)

--this is a function that is triggered by every new appearing lootmessage
function lootProxy(proxy, message)
	local creature = string.match(message, "(.+):") --this stores the first part of the lootmessage "Loot of <monster>"
	local loot = string.match(message, ": (.+)") --this stores everything after the first part of the lootmessage "a gold coin, meat, 2 worms"
	
	local itemfound = false	--we set this value false to see if it changes after the following loop
	
	--this loops through every value in the table items from above
	for _, item in ipairs(items) do
		--this checks if any of the items within the items table matches to an item in the lootmessage, and sets itemfound true if so
		if (loot:lower():match(item:lower())) then
			itemfound = true
			break --we break (stop) the loop as we have already found a match
		end
	end
	
	--this checks if an item was found, and prints an orange lootmessage into the channel if so
	if (itemfound) then
		channel:SendOrangeMessage(creature, loot)
	--this turns true for every other condition (e.g. no item was found), and prints a yellow lootmessage into the channel if so
	else
		channel:SendYellowMessage(creature, loot)
	end
end

--this continuously waits for a new lootmessage to appear, executing the function lootProxy upon receiving
LootMessageProxy.OnReceive("lootProxy", lootProxy)